﻿using CustomComponents.Exceptions;
using CustomComponents.Helpers;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;

namespace KOP_Labs
{
    public partial class CustomTree : UserControl
    {
        public CustomTree()
        {
            InitializeComponent();
        }
        
        public List<string>? Hierarcy {  get; set; }
        public void Clear()
        {
            treeView.Nodes.Clear();
        }

        public void AddNode<T>(T obj) {
            if (Hierarcy == null)
            {
                throw new HierarcyNullException("Иерархия не задана");
            }
            if (obj == null)
            {
                throw new ArgumentNullException("Объект не найден");
            }
            var values = GetValuesVithStructure(obj);
            var treeNode = new TreeNode();
            AddNodesToTreeNode(treeView.Nodes, treeNode.Nodes);
            var nodes = AddElementsToParent(values, treeNode);
            treeView.Nodes.Clear();
            AddNodesToTreeNode(nodes.Nodes, treeView.Nodes);
        }
        private void AddNodesToTreeNode(TreeNodeCollection fromNodeCollection, TreeNodeCollection toNodeCollection)
        {
            for (int i = 0; i < fromNodeCollection.Count; i++)
            {
                var node = fromNodeCollection[i].Clone() as TreeNode;
                toNodeCollection.Add(node);
            }
        }
        private Dictionary<string, (object, bool)> GetValuesVithStructure<T>(T obj) 
        {
            PropertyInfo[]? properties = obj?.GetType().GetProperties();
            var dict = new Dictionary<string, (object, bool)>();
            foreach (var elem in Hierarcy!) {
                PropertyInfo? prop = properties?.Single(prop => prop.Name == elem);
                if (prop == null)
                {
                    throw new PropertyNullException(nameof(prop));
                }
                var atr = prop.GetCustomAttributes()?.SingleOrDefault(atr => atr is AlwaysCreateAttribute);
                dict[elem] = (prop.GetValue(obj)!, atr == null ? false : true);
            }
            return dict;
        }
        private TreeNode AddElementsToParent(Dictionary<string, (object, bool)> elements, TreeNode parent)
        {
            if (elements.Count == 0)
            {
                return parent;
            }
            var firstElem = elements.First();
            var child = parent.Nodes
                .Cast<TreeNode>()
                .SingleOrDefault(node => node.Text == (string)firstElem.Value.Item1);

            if (child != null && !firstElem.Value.Item2)
            {
                elements.Remove(firstElem.Key);
                return AddElementsToParent(elements, child).Parent;
            }
            else
            {
                var newChild = new TreeNode(firstElem.Value.Item1?.ToString() ?? string.Empty);

                newChild.Tag = new Tuple<string, object>(firstElem.Key, firstElem.Value.Item1!);

                elements.Remove(firstElem.Key);
                parent.Nodes.Add(AddElementsToParent(elements, newChild));
                return parent;
            }
        }
        public T GetSelectedNode<T>()
            where T : new()
        {
            if (Hierarcy == null)
            {
                throw new HierarcyNullException("Иерархия не задана");
            }
            if (treeView.SelectedNode == null)
            {
                return new T();
            }
            return GetNode(new T(), treeView.SelectedNode);
        }
        private T GetNode<T>(T obj, TreeNode node)
        {
            if (node != null && node.Tag != null)
            {
                var tag = node.Tag as Tuple<string, object>;

                var property = obj?.GetType().GetProperty(tag!.Item1);
                property?.SetValue(obj, Convert.ChangeType(tag!.Item2, property.PropertyType));

                GetNode(obj, node.Parent);
            }
            return obj;
        }
    }
}
